"""
Contains utility functions for mity modules.
"""

import logging
import os
import subprocess
import sys
from glob import glob
from typing import Optional, Tuple
import pysam


class MityUtil:
    """
    Contains utility functions for mity modules.
    """

    MITY_DIR = "mitylib"
    REF_DIR = "reference"
    ANNOT_DIR = "annot"

    @staticmethod
    def get_mity_dir():
        """
        Get the directory path of the Mity library.

        Returns:
            str: The path to the Mity library directory.
        """
        return os.path.dirname(sys.modules["mitylib"].__file__)

    @staticmethod
    def tabix(bgzipped_file: str) -> None:
        """
        Generate a tabix index for a bgzipped file.

        Parameters:
            bgzipped_file (str): The path to a bgzip compressed file.

        Returns:
            None
        """
        tabix_call = "tabix -f " + bgzipped_file
        logging.debug(tabix_call)
        subprocess.run(tabix_call, shell=True, check=False)

    @staticmethod
    def select_reference_fasta(
        reference: str, custom_reference_fa: Optional[str] = None
    ) -> str:
        """
        Select the reference genome fasta file.

        Parameters:
            reference (str): One of the inbuilt reference genomes: hs37d5, hg19, hg38, mm10.
            custom_reference_fa (str, optional): The path to a custom reference genome, or None.

        Returns:
            str: The path to the selected reference genome fasta file.
        """
        if custom_reference_fa is not None:
            if not os.path.exists(custom_reference_fa):
                raise FileNotFoundError(
                    f"--custom-reference-fasta file: {custom_reference_fa} cannot be found."
                )
            return custom_reference_fa

        ref_dir = os.path.join(MityUtil.get_mity_dir(), MityUtil.REF_DIR)
        res = glob(f"{ref_dir}/{reference}.*.fa")
        logging.debug(",".join(res))
        assert len(res) == 1

        return res[0]

    @staticmethod
    def select_reference_genome(
        reference: str, custom_reference_genome: Optional[str] = None
    ) -> str:
        """
        Select the reference genome .genome file.

        Parameters:
            reference (str): One of the inbuilt reference genomes: hs37d5, hg19, hg38, mm10.
            custom_reference_genome: The path to a custom reference .genome file, or None.

        Returns:
            str: The path to the selected reference .genome file.
        """
        if custom_reference_genome is not None:
            if not os.path.exists(custom_reference_genome):
                raise FileNotFoundError(
                    f"--custom-reference-genome file: {custom_reference_genome} cannot be found."
                )
            return custom_reference_genome

        ref_dir = os.path.join(MityUtil.get_mity_dir(), MityUtil.REF_DIR)
        logging.debug("Looking for .genome file in %s", ref_dir)
        res = glob(f"{ref_dir}/{reference}.genome")
        logging.debug(",".join(res))
        assert len(res) == 1
        return res[0]

    @staticmethod
    def vcf_get_mt_contig(vcf: str) -> Tuple[str, Optional[int]]:
        """
        Get the mitochondrial contig name and length from a VCF file.

        Parameters:
            vcf (str): Path to a VCF file.

        Returns:
            tuple: A tuple of contig name as a str and length as an int.
        """
        r = pysam.VariantFile(vcf, "r")
        chroms = r.header.contigs
        mito_contig_intersection = set(["MT", "chrM"]).intersection(chroms)

        assert len(mito_contig_intersection) == 1

        mito_contig = "".join(mito_contig_intersection)

        mt_contig_name = r.header.contigs[mito_contig].name
        mt_contig_length = r.header.contigs[mito_contig].length

        return (mt_contig_name, mt_contig_length)

    @staticmethod
    def get_annot_file(annotation_file_path: str):
        """
        Get the path to an annotation file.

        Parameters:
            annotation_file_path (str): The name of the annotation file.

        Returns:
            str: The path to the annotation file.
        """
        mitylibdir = MityUtil.get_mity_dir()
        path = os.path.join(mitylibdir, MityUtil.ANNOT_DIR, annotation_file_path)
        assert os.path.exists(path)
        return path

    @staticmethod
    def make_prefix(vcf_path: str):
        """
        Make a prefix based on the input vcf path. This handles vcf files from
        previous steps of mity. e.g. from call to normalise, etc.

        Format of MITY output filenames:
            prefix.mity.call.vcf.gz
            prefix.mity.normalise.vcf.gz
            prefix.mity.merge.vcf.gz
            prefix.report.xlsx
        """

        prefix = (
            os.path.basename(vcf_path)
            .replace(".mity", "")
            .replace(".call", "")
            .replace(".normalise", "")
            .replace(".merge", "")
            .replace(".report", "")
            .replace(".vcf.gz", "")
        )

        return prefix

    @staticmethod
    def gsort(input_path: str, output_path: str, genome: str):
        """
        Run gsort.
        """
        gsort_cmd = f"gsort {input_path} {genome} | bgzip -cf > {output_path}"
        subprocess.run(gsort_cmd, shell=True, check=False)
        MityUtil.tabix(output_path)



"""
Mity: a sensitive variant analysis pipeline optimised for WGS data



Usage: See the online manual for details: http://github.com/KCCG/mity
Authors: Clare Puttick, Mark Cowley, Trent Zeng, Christian Fares
License: MIT
"""

import argparse
import logging
import os

from mitylib import call, normalise, report, merge, util
from ._version import __version__


usage = __doc__.split("\n\n\n", maxsplit=1)
usage[-1] += "Version: " + __version__

AP = argparse.ArgumentParser(
    description=usage[0], epilog=usage[1], formatter_class=argparse.RawTextHelpFormatter
)

# subparser    -----------------------------------------------------------------

AP_subparsers = AP.add_subparsers(help="mity sub-commands (use with -h for more info)")

# call -------------------------------------------------------------------------


def _cmd_call(args):
    """Call mitochondrial variants"""
    logging.info("mity version %s", __version__)
    logging.info("Calling mitochondrial variants")
    logging.debug("Debugging mode activated")

    genome = util.MityUtil.select_reference_genome(args.reference, args.custom_reference_genome)
    args.reference = util.MityUtil.select_reference_fasta(args.reference, args.custom_reference_fasta)

    call.Call(
        debug=args.debug,
        files=args.files,
        reference=args.reference,
        genome=genome,
        prefix=args.prefix,
        min_mq=args.min_mq,
        min_bq=args.min_bq,
        min_af=args.min_af,
        min_ac=args.min_ac,
        p=args.p,
        normalise=args.normalise,
        output_dir=args.output_dir,
        region=args.region,
        bam_list=args.bam_file_list,
        keep=args.keep,
    )


P_call = AP_subparsers.add_parser("call", help=_cmd_call.__doc__)
P_call.add_argument(
    "-d", "--debug", action="store_true", help="Enter debug mode", required=False
)
P_call.add_argument(
    "files",
    action="append",
    nargs="+",
    help="BAM / CRAM files to run the analysis on. If --bam-file-list is included, this argument is the file containing the list of bam/cram files.",
)
P_call.add_argument(
    "--reference",
    choices=["hs37d5", "hg19", "hg38", "mm10"],
    default="hs37d5",
    required=False,
    help="Reference genome version to use. Default: hs37d5",
)
P_call.add_argument(
    "--custom-reference-fasta",
    action="store",
    help="Specify custom reference fasta file",
    dest="custom_reference_fasta"
)
P_call.add_argument(
    "--custom-reference-genome",
    action="store",
    help="Specify custom reference genome file",
    dest="custom_reference_genome"
)
P_call.add_argument(
    "--prefix", action="store", help="Output files will be named with PREFIX"
)
P_call.add_argument(
    "--min-mapping-quality",
    action="store",
    type=int,
    default=30,
    help="Exclude alignments from analysis if they have a "
    "mapping quality less than MIN_MAPPING_QUALITY. "
    "Default: 30",
    dest="min_mq",
)
P_call.add_argument(
    "--min-base-quality",
    action="store",
    type=int,
    default=24,
    help="Exclude alleles from analysis if their supporting "
    "base quality is less than MIN_BASE_QUALITY. "
    "Default: 24",
    dest="min_bq",
)
P_call.add_argument(
    "--min-alternate-fraction",
    action="store",
    type=float,
    default=0.01,
    help="Require at least MIN_ALTERNATE_FRACTION "
    "observations supporting an alternate allele within "
    "a single individual in the in order to evaluate the "
    "position. Default: 0.01, range = [0,1]",
    dest="min_af",
)
P_call.add_argument(
    "--min-alternate-count",
    action="store",
    type=int,
    default=4,
    help="Require at least MIN_ALTERNATE_COUNT observations "
    "supporting an alternate allele within a single "
    "individual in order to evaluate the position. "
    "Default: 4",
    dest="min_ac",
)
P_call.add_argument(
    "--p",
    action="store",
    type=float,
    default=0.002,
    help="Minimum noise level. This is used to calculate QUAL score. "
    "Default: 0.002, range = [0,1]",
    dest="p",
)
P_call.add_argument(
    "--normalise", action="store_true", help="Run mity normalise the resulting VCF"
)
P_call.add_argument(
    "--output-dir",
    action="store",
    type=str,
    default=".",
    help="Output files will be saved in OUTPUT_DIR. " "Default: '.' ",
    dest="output_dir",
)
P_call.add_argument(
    "--region",
    action="store",
    type=str,
    default=None,
    help="Region of MT genome to call variants in. "
    "If unset will call variants in entire MT genome as specified in BAM header. "
    "Default: Entire MT genome. ",
    dest="region",
)
P_call.add_argument(
    "--bam-file-list",
    action="store_true",
    default=False,
    help="Treat the file as a text file of BAM files to be processed."
    " The path to each file should be on one row per bam file.",
    dest="bam_file_list",
)
P_call.add_argument(
    "-k",
    "--keep",
    action="store_true",
    required=False,
    help="Keep all intermediate files",
)
P_call.set_defaults(func=_cmd_call)

# normalise --------------------------------------------------------------------


def _cmd_normalise(args):
    """Normalise & FILTER mitochondrial variants"""
    logging.info("mity %s", __version__)
    logging.info("Normalising and FILTERing mitochondrial vcf.gz file")

    genome = util.MityUtil.select_reference_genome(args.reference, args.custom_reference_genome)
    args.reference = util.MityUtil.select_reference_fasta(args.reference, args.custom_reference_fasta)

    normalise.Normalise(
        debug=args.debug,
        vcf=args.vcf,
        reference_fasta=args.reference,
        genome=genome,
        output_dir=args.output_dir,
        prefix=args.prefix,
        allsamples=args.allsamples,
        keep=args.keep,
        p=args.p,
    )


P_normalise = AP_subparsers.add_parser("normalise", help=_cmd_normalise.__doc__)
P_normalise.add_argument(
    "-d", "--debug", action="store_true", help="Enter debug mode", required=False
)
P_normalise.add_argument("vcf", action="store", help="vcf.gz file from running mity")
P_normalise.add_argument(
    "--output-dir",
    action="store",
    type=str,
    default=".",
    help="Output files will be saved in OUTPUT_DIR. " "Default: '.' ",
    dest="output_dir",
)
P_normalise.add_argument(
    "--prefix", action="store", help="Output files will be named with PREFIX"
)
P_normalise.add_argument(
    "--allsamples",
    action="store_true",
    required=False,
    help="PASS in the filter requires all samples to pass instead of just one",
)
P_normalise.add_argument(
    "-k",
    "--keep",
    action="store_true",
    required=False,
    help="Keep all intermediate files",
)
P_normalise.add_argument(
    "--p",
    action="store",
    type=float,
    default=0.002,
    help="Minimum noise level. This is used to calculate QUAL score"
    "Default: 0.002, range = [0,1]",
    dest="p",
)
P_normalise.add_argument(
    "--reference",
    choices=["hs37d5", "hg19", "hg38", "mm10"],
    default="hs37d5",
    required=False,
    help="Reference genome version to use. default: hs37d5",
)
P_normalise.add_argument(
    "--custom-reference-fasta",
    action="store",
    help="Specify custom reference fasta file",
    dest="custom_reference_fasta"
)
P_normalise.add_argument(
    "--custom-reference-genome",
    action="store",
    help="Specify custom reference genome file",
    dest="custom_reference_genome"
)
P_normalise.set_defaults(func=_cmd_normalise)


# report -----------------------------------------------------------------------


def _cmd_report(args):
    """Generate mity report"""
    logging.info("mity %s", __version__)
    logging.info("Generating mity report")
    report.Report(
        debug=args.debug,
        vcfs=args.vcf,
        contig=args.contig,
        prefix=args.prefix,
        min_vaf=args.min_vaf,
        output_dir=args.output_dir,
        keep=args.keep,
        vcfanno_config=args.vcfanno_config,
        report_config=args.report_config,
    )


P_report = AP_subparsers.add_parser("report", help=_cmd_report.__doc__)
P_report.add_argument(
    "-d", "--debug", action="store_true", help="Enter debug mode", required=False
)
P_report.add_argument(
    "--prefix", action="store", help="Output files will be named with PREFIX"
)
P_report.add_argument(
    "--min_vaf",
    action="store",
    type=float,
    default=0,
    help="A variant must have at least this VAF to be included in the report. Default: "
    "0.",
)
P_report.add_argument(
    "--output-dir",
    action="store",
    type=str,
    default=".",
    help="Output files will be saved in OUTPUT_DIR. " "Default: '.' ",
    dest="output_dir",
)
P_report.add_argument(
    "vcf", action="append", nargs="+", help="mity vcf files to create a report from"
)
P_report.add_argument(
    "-k",
    "--keep",
    action="store_true",
    required=False,
    help="Keep all intermediate files",
)
P_report.add_argument(
    "--contig",
    choices=["MT", "chrM"],
    default="MT",
    required=False,
    help="Contig used for annotation purposes",
)
P_report.add_argument(
    "--custom-vcfanno-config",
    action="store",
    help="Provide a custom vcfanno-config.toml for custom annotations.",
    dest="vcfanno_config",
)
P_report.add_argument(
    "--custom-report-config",
    action="store",
    help="Provide a custom report-config.yaml for custom report generation.",
    dest="report_config",
)
P_report.set_defaults(func=_cmd_report)


# merge -----------------------------------------------------------------------


def _cmd_merge(args):
    """Merging mity VCF with nuclear VCF"""
    logging.info("mity %s", __version__)
    logging.info("mity vcf merge")

    genome = util.MityUtil.select_reference_genome(args.reference, None)

    merge.Merge(
        debug=args.debug,
        nuclear_vcf_path=args.nuclear_vcf,
        mity_vcf_path=args.mity_vcf,
        genome=genome,
        output_dir=args.output_dir,
        prefix=args.prefix,
        keep=args.keep,
    )


P_merge = AP_subparsers.add_parser("merge", help=_cmd_merge.__doc__)
P_merge.add_argument("--mity_vcf", action="store", required=True, help="mity vcf file")
P_merge.add_argument(
    "--nuclear_vcf", action="store", required=True, help="nuclear vcf file"
)
P_merge.add_argument(
    "--output-dir",
    action="store",
    type=str,
    default=".",
    help="Output files will be saved in OUTPUT_DIR. " "Default: '.' ",
    dest="output_dir",
)
P_merge.add_argument(
    "--prefix",
    action="store",
    help="Output files will be named with PREFIX. "
    "The default is to use the nuclear vcf name",
)
P_merge.add_argument(
    "--reference",
    choices=["hs37d5", "hg19", "hg38", "mm10"],
    default="hs37d5",
    required=False,
    help="reference genome version to use. default: hs37d5",
)
P_merge.add_argument(
    "-d", "--debug", action="store_true", help="Enter debug mode", required=False
)
P_merge.add_argument(
    "-k",
    "--keep",
    action="store_true",
    required=False,
    help="Keep all intermediate files",
)
P_merge.set_defaults(func=_cmd_merge)


# runall -----------------------------------------------------------------------


def _cmd_runall(args):
    """Run MITY call, normalise and report all in one go."""
    logging.info("mity %s", __version__)
    logging.info("mity runall")

    genome = util.MityUtil.select_reference_genome(args.reference, args.custom_reference_genome)
    args.reference = util.MityUtil.select_reference_fasta(args.reference, args.custom_reference_fasta)

    call.Call(
        debug=args.debug,
        files=args.files,
        reference=args.reference,
        genome=genome,
        prefix=args.prefix,
        min_mq=args.min_mq,
        min_bq=args.min_bq,
        min_af=args.min_af,
        min_ac=args.min_ac,
        p=args.p,
        # normalise flag is set to true instead of running normalise separately
        normalise=True,
        output_dir=args.output_dir,
        region=args.region,
        bam_list=args.bam_file_list,
        keep=args.keep,
    )

    logging.debug("mity call and normalise completed")

    # This makes use of the uniform naming scheme of mity command outputs.
    normalised_vcf_path = os.path.join(
        args.output_dir, args.prefix + ".normalise.vcf.gz"
    )

    logging.debug("assumed mity normalise vcf output path is: %s", normalised_vcf_path)

    # matching argparse quirk
    normalised_vcf_path = [normalised_vcf_path]

    report.Report(
        debug=args.debug,
        vcfs=normalised_vcf_path,
        contig=args.contig,
        prefix=args.prefix,
        min_vaf=args.min_vaf,
        output_dir=args.output_dir,
        keep=args.keep,
        vcfanno_config=args.vcfanno_config,
        report_config=args.report_config,
    )


P_runall = AP_subparsers.add_parser("runall", help=_cmd_runall.__doc__)
P_runall.add_argument(
    "-d", "--debug", action="store_true", help="Enter debug mode", required=False
)
P_runall.add_argument(
    "files",
    action="append",
    nargs="+",
    help="BAM / CRAM files to run the analysis on. If --bam-file-list is included, this argument is the file containing the list of bam/cram files.",
)
P_runall.add_argument(
    "--reference",
    choices=["hs37d5", "hg19", "hg38", "mm10"],
    default="hs37d5",
    required=False,
    help="Reference genome version to use. Default: hs37d5",
)
# For the runall command, we mandate that the prefix option is set. This is not
# true for regular mity call, normalise or report separately.
P_runall.add_argument(
    "--prefix",
    action="store",
    required=True,
    help="Output files will be named with PREFIX",
)
P_runall.add_argument(
    "--min-mapping-quality",
    action="store",
    type=int,
    default=30,
    help="Exclude alignments from analysis if they have a "
    "mapping quality less than MIN_MAPPING_QUALITY. "
    "Default: 30",
    dest="min_mq",
)
P_runall.add_argument(
    "--min-base-quality",
    action="store",
    type=int,
    default=24,
    help="Exclude alleles from analysis if their supporting "
    "base quality is less than MIN_BASE_QUALITY. "
    "Default: 24",
    dest="min_bq",
)
P_runall.add_argument(
    "--min-alternate-fraction",
    action="store",
    type=float,
    default=0.01,
    help="Require at least MIN_ALTERNATE_FRACTION "
    "observations supporting an alternate allele within "
    "a single individual in the in order to evaluate the "
    "position. Default: 0.01, range = [0,1]",
    dest="min_af",
)
P_runall.add_argument(
    "--min-alternate-count",
    action="store",
    type=int,
    default=4,
    help="Require at least MIN_ALTERNATE_COUNT observations "
    "supporting an alternate allele within a single "
    "individual in order to evaluate the position. "
    "Default: 4",
    dest="min_ac",
)
P_runall.add_argument(
    "--p",
    action="store",
    type=float,
    default=0.002,
    help="Minimum noise level. This is used to calculate QUAL score. "
    "Default: 0.002, range = [0,1]",
    dest="p",
)
P_runall.add_argument(
    "--output-dir",
    action="store",
    type=str,
    default=".",
    help="Output files will be saved in OUTPUT_DIR. " "Default: '.' ",
    dest="output_dir",
)
P_runall.add_argument(
    "--region",
    action="store",
    type=str,
    default=None,
    help="Region of MT genome to call variants in. "
    "If unset will call variants in entire MT genome as specified in BAM header. "
    "Default: Entire MT genome. ",
    dest="region",
)
P_runall.add_argument(
    "--bam-file-list",
    action="store_true",
    default=False,
    help="Treat the file as a text file of BAM files to be processed."
    " The path to each file should be on one row per bam file.",
    dest="bam_file_list",
)
P_runall.add_argument(
    "-k",
    "--keep",
    action="store_true",
    required=False,
    help="Keep all intermediate files",
)
P_runall.add_argument(
    "--min_vaf",
    action="store",
    type=float,
    default=0,
    help="A variant must have at least this VAF to be included in the report. Default: "
    "0.",
)
P_runall.add_argument(
    "--contig",
    choices=["MT", "chrM"],
    default="MT",
    required=False,
    help="Contig used for annotation purposes",
)
P_runall.add_argument(
    "--custom-vcfanno-config",
    action="store",
    help="Provide a custom vcfanno-config.toml for custom annotations.",
    dest="vcfanno_config",
)
P_runall.add_argument(
    "--custom-report-config",
    action="store",
    help="Provide a custom report-config.yaml for custom report generation.",
    dest="report_config",
)
P_runall.add_argument(
    "--custom-reference-fasta",
    action="store",
    help="Specify custom reference fasta file",
    dest="custom_reference_fasta"
)
P_runall.add_argument(
    "--custom-reference-genome",
    action="store",
    help="Specify custom reference genome file",
    dest="custom_reference_genome"
)
P_runall.set_defaults(func=_cmd_runall)


# version ----------------------------------------------------------------------


def print_version(_args):
    """Display this program's version."""
    print(__version__)


P_version = AP_subparsers.add_parser("version", help=print_version.__doc__)
P_version.set_defaults(func=print_version)

# parse_args -------------------------------------------------------------------


def parse_args(args=None):
    """Parse the command line."""
    return AP.parse_args(args=args)



"""Normalise mity VCF."""

import logging
import os.path
from math import isinf
from scipy.stats import binom
import numpy as np
import pysam
import pysam.bcftools
from mitylib.util import MityUtil


# LOGGING
logger = logging.getLogger(__name__)


class Normalise:
    """
    Mity normalise.
    """

    P_VAL = 0.002
    SB_RANGE_LO = 0.1
    SB_RANGE_HI = 0.9
    MIN_MQMR = 30
    MIN_AQR = 20
    MIN_DP = 15
    BLACKLIST = list(range(302, 319)) + list(range(3105, 3109))

    def __init__(
        self,
        debug,
        vcf,
        reference_fasta,
        genome,
        output_dir=".",
        prefix=None,
        allsamples=False,
        keep=False,
        p=P_VAL,
    ):
        self.debug = debug
        self.vcf = vcf
        self.reference_fasta = reference_fasta
        self.genome = genome
        self.output_dir = output_dir
        self.prefix = prefix
        self.allsamples = allsamples
        self.keep = keep
        self.p = p

        self.bcftools_norm_obj = None
        self.filtered_vcf_obj = None

        self.bcftools_norm_path = ""
        self.filtered_vcf_path = ""
        self.normalised_vcf_path = ""

        self.run()

    def run(self):
        """
        Run mity normalise.
        """
        if self.debug:
            logger.setLevel(logging.DEBUG)
            logger.debug("Entered debug mode.")
        else:
            logger.setLevel(logging.INFO)

        self.set_strings()
        self.run_bcftools_norm()
        self.run_filtering()

        MityUtil.gsort(self.filtered_vcf_path, self.normalised_vcf_path, self.genome)

        self.remove_intermediate_files()

    def run_filtering(self):
        """
        Makes pysam VariantFile objects and runs filtering for each variant.
        """
        self.bcftools_norm_obj = pysam.VariantFile(self.bcftools_norm_path)
        self.filtered_vcf_obj = pysam.VariantFile(
            self.filtered_vcf_path, "w", header=self.add_headers()
        )

        for variant in self.bcftools_norm_obj.fetch():
            variant = self.add_info_values(variant)
            variant = self.add_filter(variant)

            self.filtered_vcf_obj.write(variant)

        self.bcftools_norm_obj.close()
        self.filtered_vcf_obj.close()

    def run_bcftools_norm(self):
        """
        Run bcftools norm.
        """
        # The "-o" option doesn't get passed to pysam's bcftools wrapper (as of 31-08-2023)
        # so we write to a separate file manually.
        with open(self.bcftools_norm_path, "w") as f:
            print(
                pysam.bcftools.norm("-f", self.reference_fasta, "-m-both", self.vcf),
                end="",
                file=f,
            )

    def set_strings(self):
        """
        Sets:
            prefix
            normalised_vcf
            filtered_vcf
        """
        if self.prefix is None:
            self.prefix = MityUtil.make_prefix(self.vcf)

        self.bcftools_norm_path = os.path.join(
            self.output_dir, self.prefix + ".bcftools.norm.vcf.gz"
        )
        self.filtered_vcf_path = self.vcf.replace(".vcf.gz", ".filtered.vcf")
        self.normalised_vcf_path = os.path.join(
            self.output_dir, self.prefix + ".normalise.vcf.gz"
        )

    def add_headers(self):
        """
        Return new headers for mity normalise vcf output.
        """
        new_header = self.bcftools_norm_obj.header

        # fitler headers
        new_header.filters.add(
            "FAIL",
            None,
            None,
            "Variant fails filter in at least one or all samples (depending on filter option)",
        )

        # info headers
        new_header.info.add(
            "SBR",
            number=1,
            type="Float",
            description="For all alleles RO > 15 and (SBR > 0.9 or SBR < 0.1)",
        )
        new_header.info.add(
            "SBA",
            number=1,
            type="Float",
            description="For all alleles AO > 15 and (SBA > 0.9 or SBA < 0.1)",
        )

        # format headers
        new_header.formats.add(
            "AQA",
            number="A",
            type="Float",
            description="Average base quality of the alternate reads, AQA=QA/AO",
        )
        new_header.formats.add(
            "AQR",
            number="A",
            type="Float",
            description="Average base quality of the reference reads, AQR=QR/RO",
        )
        new_header.formats.add(
            "VAF",
            number="A",
            type="Float",
            description="Allele frequency in the range (0,1] - the ratio of the number of alternate reads to reference reads",
        )
        new_header.formats.add(
            "q",
            number="A",
            type="Float",
            description="Phred scaled binomial probability of seeing AO reads from DP, assuming a noise floor of p=0.002. ",
        )
        new_header.formats.add(
            "tier",
            number="A",
            type="Integer",
            description="Custom variant tier (TODO: change description)",
        )

        # format filter headers
        new_header.formats.add(
            "POS_filter",
            number="A",
            type="Integer",
            description="Variant falls in the blacklist of positions: MT:302-319, MT:3105-3108",
        )
        new_header.formats.add(
            "SBR_filter",
            number="A",
            type="Integer",
            description="For all alleles RO > 15 and (SBR > 0.9 or SBR < 0.1)",
        )
        new_header.formats.add(
            "SBA_filter",
            number="A",
            type="Integer",
            description="For all alleles AO > 15 and (SBA > 0.9 or SBA < 0.1)",
        )
        new_header.formats.add(
            "MQMR_filter",
            number="A",
            type="Integer",
            description="For all alleles MQMR<30",
        )
        new_header.formats.add(
            "AQR_filter",
            number="A",
            type="Integer",
            description="For all alleles AQR<20",
        )

        return new_header

    def mity_qual(self, AO, DP, p):
        """
        Calculate the Phred-scaled quality score for a genetic variant.

        Parameters:
            - AO (int): Number of alternative reads.
            - DP (int): Total read depth.

        Returns:
            - float: Phred-scaled quality score.
        """
        q = 0.0
        AO = int(AO)
        DP = int(DP)
        if AO > 0 and DP > 0:
            if DP == AO:
                # then penalise homoplasmic variants with low numbers of reads
                DP = DP + 1
            # v1: the cdf implementation caps at 159.55, due to the cdf capping at 0.9999999999999999
            # q = round(abs(-10 * log10(1 - binom.cdf(AO, DP, p))), 2)
            # v2: the logsf implementation is identical to cdf for low values and continues to scale up
            with np.errstate(divide="ignore"):
                q = float(round(-4.342945 * binom.logsf(AO, DP, p), 2))
        if isinf(q):
            # this is the maximum q that we observed looking at homoplastic variants around ~129x depth
            q = float(3220)
        return q

    def add_filter(self, variant):
        """
        Adds filter to variant, and filters for samples.

        Parameters:
            - variant (VariantRecord): Variant to add filters to

        Returns:
            - VariantRecord: Variant with filters added
        """
        # filter dictionary
        # each value represents the number of samples that pass each field
        # starts with all samples passing
        num_samples = len(variant.samples.values())
        filter_dict = {
            "SBR": num_samples,
            "MQMR": num_samples,
            "AQR": num_samples,
            "SBA": num_samples,
        }

        # filtering
        pass_flag = True
        pos_flag = True
        if variant.pos in self.BLACKLIST:
            pos_flag = False
            pass_flag = False

        # filtering samples
        for sample in variant.samples.values():
            # adding to format
            # AQR
            if int(sample["RO"]) > 0:
                AQR = float(sample["QR"]) / float(sample["RO"])
            else:
                AQR = 0

            sample["AQR"] = AQR

            # VAF
            if sample["DP"] != 0:
                VAF = round(float(sample["AO"][0]) / float(sample["DP"]), 4)
            else:
                VAF = 0

            sample["VAF"] = VAF

            # mity quality
            q = self.mity_qual(sample["AO"][0], sample["DP"], self.p)
            sample["q"] = q

            # AQA
            if sample["AO"][0] > 0:
                AQA = float(round(sample["QA"][0] / sample["AO"][0], 3))
            else:
                AQA = float(0)

            sample["AQA"] = AQA

            # tier (originally from mity report)
            if float(VAF) >= 0.01:
                tier = 1
            elif float(VAF) < 0.01 and float(sample["AO"][0]) > 10:
                tier = 2
            else:
                tier = 3

            sample["tier"] = tier

            # set tests to "PASS", i.e. 1 first
            # sidenote: it would be nice to use TRUE/FALSE but the format field does
            # not support boolean values
            sample["POS_filter"] = 1 if pos_flag else 0
            sample["SBR_filter"] = 1
            sample["SBA_filter"] = 1
            sample["MQMR_filter"] = 1
            sample["AQR_filter"] = 1

            # testing
            if sample["RO"] > self.MIN_DP:
                if not self.SB_RANGE_LO <= variant.info["SBR"] <= self.SB_RANGE_HI:
                    filter_dict["SBR"] -= 1
                    sample["SBR_filter"] = 0
                if variant.info["MQMR"] < self.MIN_MQMR:
                    filter_dict["MQMR"] -= 1
                    sample["MQMR_filter"] = 0
                if AQR < self.MIN_AQR:
                    filter_dict["AQR"] -= 1
                    sample["AQR_filter"] = 0
            if sample["AO"][0] > self.MIN_DP:
                if not self.SB_RANGE_LO <= variant.info["SBA"] <= self.SB_RANGE_HI:
                    filter_dict["SBA"] -= 1
                    sample["SBA_filter"] = 0

        # all samples must pass each test
        if self.allsamples:
            for num_passing_samples in filter_dict.values():
                if num_passing_samples != num_samples:
                    pass_flag = False
                    break

        # only one sample has to pass each test
        else:
            for num_passing_samples in filter_dict.values():
                if num_passing_samples == 0:
                    pass_flag = False
                    break

        # setting passing filter
        if pass_flag:
            variant.filter.add("PASS")
        else:
            variant.filter.add("FAIL")

        return variant

    def add_info_values(self, variant):
        """
        Adds new and calculated INFO values to a variant.
        """
        # adding to INFO field

        # we want to caclulate SBR and add to INFO
        # SBR = strand bias of reference
        # SBR = SRF/(SRF+SRR)
        SRF = variant.info["SRF"]
        SRR = variant.info["SRR"]
        # to avoid dividing by zero:
        if SRF > 0 or SRR > 0:
            SBR = round(SRF / (SRF + SRR), 3)
        else:
            SBR = 0

        # we want to calculate SBA and add to INFO
        # SBA = strand bias of alternate
        # i.e. SBA = SAF/(SAF+SAR)

        # >>> variant.info["SAF"]
        # >>> (8505,)
        # so we use [0] to get the first value, and we can assume that there is only one value
        SAF = variant.info["SAF"][0]
        SAR = variant.info["SAR"][0]

        SBA = round(SAF / (SAF + SAR), 3)

        variant.info.update({"SBR": float(SBR), "SBA": float(SBA)})

        return variant

    def remove_intermediate_files(self):
        """
        Remove intermediate files.
        """
        if not self.keep:
            os.remove(self.filtered_vcf_path)
            os.remove(self.bcftools_norm_path)



import setuptools

with open("mitylib/_version.py", "r") as fh:
    version = fh.read().replace("__version__ = ", "").strip('""\n')

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mitywgs",
    version=version,
    description="A sensitive Mitochondrial variant detection pipeline from WGS data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/KCCG/mity",
    author="Mark Cowley",
    author_email="mcowley@ccia.org.au",
    license="MIT",
    license_files = ["LICEN[CS]E*"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Healthcare Industry",
        "License :: Free for non-commercial use",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="mitochondrial DNA genomics variant SNV INDEL",
    project_urls={
        "Documentation": "https://github.com/KCCG/mity/",
        "Source": "https://github.com/KCCG/mity/",
        "Funding": "http://garvan.org.au/kccg",
    },
    packages=setuptools.find_packages(),
    install_requires=["pysam", "pandas", "xlsxwriter", "scipy", "pyyaml"],
    python_requires=">=3.10",
    package_data={"mitylib": ["annot_mt/*", "annot_chrm/*", "reference/*", "config/*"]},
    include_package_data=True,
    scripts=["mitylib/mity"],
)



"""
Adds annotations and generates mity excel report from VCF file.
"""

import logging
import os.path
import subprocess
from typing import Dict, Optional
import pysam
import pandas
import yaml


from mitylib.util import MityUtil

# LOGGING
logger = logging.getLogger(__name__)


class Vep:
    """
    Provides methods to handle vepped files.
    """

    def __init__(self, vcf_obj):
        self.vcf_obj = vcf_obj
        self.vepped = self.check_vepped()

        if self.vepped:
            self.vep_keys = self.generate_vep_keys()
            self.vep_excel_headers = self.generate_vep_excel_headers()

    def get_vep_name(self, name):
        """
        Capitalises vep name, replaces _ with <space> and adds "VEP" to the end.
        """
        name = name.upper()
        name = name.replace("_", " ")
        name = name + " VEP"

        return name

    def generate_vep_excel_headers(self):
        """
        Generates vep excel headers based on the vep keys.
        """
        vep_excel_headers = ["HIGHEST IMPACT VEP"]
        for name in self.vep_keys:
            vep_excel_headers.append(self.get_vep_name(name))

        return vep_excel_headers

    def find_highest_impact(self, impacts):
        """
        Returns the highest impact in the list of consequences. Raises error if
        none of the consequences match.
        """
        if "HIGH" in impacts:
            return "HIGH"
        if "MODERATE" in impacts:
            return "MODERATE"
        if "MODIFIER" in impacts:
            return "MODIFIER"
        if "LOW" in impacts:
            return "LOW"

        logging.warning("Unknown SO term.")

        return ""

    def get_vep_values(self, variant):
        """
        Takes a string from VEP consequences/impacts in the form:
            impact value | impact value | ... |, |||, |||

        And peforms the following:
            - removes annotation "line" if the consequence has "stream" in it,
              e.g. "upstream"
            - concatenates remaining line fields with ";"

        Types:
            variant.info["CSQ"]: Tuple of form (a|b|c|..., a|b|c|..., a|b|c|...)

        Example:
            vep_keys = [ IMPACT, Consequence, field_1, field_2, field_3 ]

            list(variant.info["CSQ"] = [
                HIGH        | something_else    | a | b | c,
                LOW         | upstream_variant  | d | e | f,
                MODIFIER    | something_else    | g | h | i
            ]

            vep_dict = {
                HIGHEST IMPACT VEP  : HIGH,
                CONSEQUENCE VEP     : something_else;something_else,
                FIELD 1 VEP         : a;g,
                FIELD 2 VEP         : b;h,
                FIELD 3 VEP         : c;i
            }
        """

        vep_dict = {}
        for name in self.vep_excel_headers:
            vep_dict[name] = []

        impacts = []

        for line in list(variant.info["CSQ"]):
            split_line = line.split("|")

            # NOTE: rename IMPACT if the name changes to something else
            impact = split_line[self.vep_keys.index("IMPACT")]
            if "stream" in impact:
                continue

            impacts.append(impact)

            for i, value in enumerate(split_line):
                if value:
                    vep_dict[self.vep_excel_headers[i + 1]].append(value)

        vep_dict["HIGHEST IMPACT VEP"] = self.find_highest_impact(impacts)

        # NOTE: final string format/join for excel output can be changed here
        for key, value in vep_dict.items():
            if isinstance(value, list):
                vep_dict[key] = ";".join(value)

        return vep_dict

    def check_vepped(self):
        """
        Checks whether a file has been vepped by looking for VEP in the vcf
        header. The line should look something like this:
            ##VEP="v110" time="2023-10-04 00:20:30" cache="/net/isilonP...
        """
        if "VEP" in str(self.vcf_obj.header):
            return True
        return False

    def generate_vep_keys(self):
        """
        Example header INFO line:
            ##INFO=<ID=CSQ,Number=.,Type=String,Description=
            "Consequence annotations from Ensembl VEP. Format:
            Allele|Consequence|IMPACT|SYMBOL|Gene|Feature_type|Feature|...

        Returns a list of keys, i.e. ["Allele", "Consequence", ...]

        NOTE: This method is hard coded based on the format of the description.
        """
        for header in self.vcf_obj.header.records:
            if header.type == "INFO" and "CSQ" in str(header) and header.attrs:
                description = header.attrs[-2][1].strip('"')

                # change this line if the description text or format changes
                description = description.replace(
                    "Consequence annotations from Ensembl VEP. Format:", ""
                )

                keys = description.split("|")
                return keys

        return None


class SingleReport:
    """
    Handles generating a mity report for one VCF file.
    """

    def __init__(
        self, vcf_path, min_vaf, keep, vcfanno_base_path, vcfanno_config, report_config
    ) -> None:
        self.min_vaf = min_vaf
        self.keep = keep

        self.vcf_path = vcf_path
        self.vcf_obj = pysam.VariantFile(vcf_path)

        self.annot_vcf_path = None
        self.annot_vcf_obj = None

        self.vcfanno_base_path = vcfanno_base_path
        self.vcfanno_config = vcfanno_config
        self.report_config = report_config

        # excel and vcf headers
        with open(self.report_config, "r", encoding="utf-8") as report_config_file:
            opened_report_config = yaml.safe_load(report_config_file)

        self.excel_headers = opened_report_config["excel_headers"]
        self.vcf_headers = opened_report_config["vcf_headers"]

        self.excel_table: Dict[str, list[str]] = {}
        self.df = None

        self.vep: Optional[Vep] = None

        self.run()

    def run(self):
        """
        Run SingleReport.
        """

        self.vcfanno_call()
        self.vep = Vep(self.annot_vcf_obj)
        self.make_table()

        if not self.keep:
            # remove vcfanno annotated vcf
            os.remove(self.annot_vcf_path)

    def get_df(self):
        """
        Return a pandas dataframe from the generated excel table.
        """
        all_excel_headers = self.excel_headers
        if self.vep.vepped:
            all_excel_headers += self.vep.vep_excel_headers
        return pandas.DataFrame(self.excel_table, columns=all_excel_headers)

    def make_hgvs(self, pos, ref, alt):
        """
        Creates HGVS syntax used in the HGVS column of the table/excel spreadsheet.
        """
        if len(alt) > 1 or len(ref) > 1:
            # this is an indel
            if len(ref) > len(alt):
                # this is a del
                delet = ref[1:]
                if len(delet) == 1:
                    hgvs_pos = int(pos) + 1
                elif len(delet) > 1:
                    hvgs_pos_start = int(pos) + 1
                    hvgs_pos_end = int(pos) + len(delet)
                    hgvs_pos = str(hvgs_pos_start) + "_" + str(hvgs_pos_end)
                hgvs = "m." + str(hgvs_pos) + "del"

            else:
                # this is an ins
                ins = alt[1:]
                if len(ins) == 1:
                    hgvs_pos = int(pos) + 1
                elif len(ins) > 1:
                    hvgs_pos_start = int(pos) + 1
                    hvgs_pos_end = int(pos) + len(ins)
                    hgvs_pos = str(hvgs_pos_start) + "_" + str(hvgs_pos_end)
                hgvs = "m." + str(hgvs_pos) + "ins"

        else:
            # this is a SNP
            hgvs = "m." + str(pos) + str(ref) + ">" + str(alt)
        return hgvs

    def vcfanno_call(self):
        """
        Calls vcfanno to annotate the output of mity normalise with the relevant
        annotations based on vcfanno-config.toml in mitylib/
        """
        logger.debug("Running vcfanno...")

        # annotated_file name
        annotated_file = self.vcf_path.replace(".vcf.gz", ".mity.annotated.vcf")
        base_path_arg = (
            f"-base-path {self.vcfanno_base_path}" if self.vcfanno_base_path else ""
        )

        # vcfanno call
        vcfanno_cmd = f"vcfanno -p 4 {base_path_arg} {self.vcfanno_config} {self.vcf_path} > {annotated_file}"
        res = subprocess.run(
            vcfanno_cmd,
            shell=True,
            check=False,
            # capture the output of the vcfanno command since it tends to produce
            # long warning messages, output shown in --debug mode
            capture_output=True,
            text=True,
        )

        logger.debug("vcfanno output:")
        logger.debug(res.stdout)

        self.annot_vcf_path = annotated_file
        self.annot_vcf_obj = pysam.VariantFile(annotated_file)

    def make_info_string(self, variant):
        """
        Takes the INFO fields from the variant and recreates the format as it
        appears in a vcf file. i.e.

        FIELD=X;FIELD=Y;etc
        """
        info_field_array = []
        for key, value in variant.info.items():
            if isinstance(value, tuple):
                value = value[0]
            info_field_array.append(f"{key}={value}")
        info_field_string = ";".join(info_field_array)
        return info_field_string

    def make_format_string(self, sample):
        """
        Takes the FORMAT fields from the sample and recreates the format as it
        appears in a vcf file. i.e.

        FIELD=X;FIELD=Y;etc
        """
        format_field_array = []
        for value in sample.values():
            if isinstance(value, tuple):
                value = value[0]
            format_field_array.append(str(value))
        format_field_string = ":".join(format_field_array)
        return format_field_string

    def clean_string(self, s: str) -> str:
        """
        Removes the following characters from a string:
            "
            '
            ()

        Used to output text from annotation sources.
        """
        if isinstance(s, tuple):
            s = s[0]

        s = str(s)
        s = s.replace('"', "").replace("(", "").replace(")", "")

        return s

    def make_table(self):
        """
        Takes a vcfanno annotated vcf file and returns a formatted dictionary
        with relevant information.

        The vcf and excel header names are hardcoded in report-config.yaml.
        """

        for header in self.excel_headers:
            self.excel_table[header] = []

        # check if VCF file is vepped, adds relevant VEP headers

        if self.vep.vepped:
            for header in self.vep.vep_excel_headers:
                self.excel_table[header] = []

        num_samples = len(self.annot_vcf_obj.header.samples)

        for variant in self.annot_vcf_obj.fetch():
            cohort_count = 0
            info_string = self.make_info_string(variant)

            # samples
            for sample in variant.samples.values():
                # skip sample if the VAF is too low
                if float(sample["VAF"][0]) <= float(self.min_vaf):
                    continue

                cohort_count += 1

                self.excel_table["SAMPLE"].append(sample.name)
                self.excel_table["HGVS"].append(
                    self.make_hgvs(variant.pos, variant.ref, variant.alts[0])
                )

                self.excel_table["CHR"].append(variant.chrom)
                self.excel_table["POS"].append(variant.pos)
                self.excel_table["REF"].append(variant.ref)

                # NOTE: assumes only one ALT as the vcf should be normalised
                self.excel_table["ALT"].append(variant.alts[0])
                self.excel_table["QUAL"].append(variant.qual)
                self.excel_table["FILTER"].append(variant.filter.keys()[0])

                self.excel_table["INFO"].append(info_string)
                self.excel_table["FORMAT"].append(self.make_format_string(sample))

                # vcf_headers: info
                for vcf_header, excel_header in self.vcf_headers["info"].items():
                    if vcf_header in variant.info.keys():
                        self.excel_table[excel_header].append(
                            self.clean_string(variant.info[vcf_header])
                        )
                    else:
                        self.excel_table[excel_header].append(".")

                # vcf_headers: annotations
                for vcf_header, excel_header in self.vcf_headers["annotations"].items():
                    if vcf_header in variant.info.keys():
                        self.excel_table[excel_header].append(
                            self.clean_string(variant.info[vcf_header])
                        )
                    else:
                        self.excel_table[excel_header].append(".")

                # vcf_headers: format
                for vcf_header, excel_header in self.vcf_headers["format"].items():
                    if vcf_header in sample.keys():
                        self.excel_table[excel_header].append(
                            self.clean_string(sample[vcf_header])
                        )
                    else:
                        self.excel_table[excel_header].append(".")

                # calculate ALLELE FREQUENCY MITOMAP
                if "Genbank_frequency_mitomap" in sample.keys():
                    allele_frequency_mitomap = round(
                        float(sample["GenBank_frequency_mitomap"]) / 32059, 3
                    )
                else:
                    allele_frequency_mitomap = "."

                self.excel_table["ALLELE FREQUENCY MITOMAP"].append(
                    allele_frequency_mitomap
                )

            # fill in cohort count
            cohort_frequency = float(cohort_count / num_samples)
            for _ in range(cohort_count):
                self.excel_table["COHORT COUNT"].append(cohort_count)
                self.excel_table["COHORT FREQUENCY"].append(cohort_frequency)

            # fill in VEP impacts if vepped
            if self.vep.vepped:
                self.add_vep_impacts(variant, cohort_count)

    def add_vep_impacts(self, variant, cohort_count: int) -> None:
        """
        Adds vep impacts for a variant.
        """
        if not self.vep:
            raise RuntimeError("Vep not initialised, something likely went wrong with vcfanno")

        variant_impacts = self.vep.get_vep_values(variant)
        for _ in range(cohort_count):
            for vep_header in self.vep.vep_excel_headers:
                self.excel_table[vep_header].append(variant_impacts[vep_header])


class Report:
    """
    Runs mity report.
    """

    def __init__(
        self,
        debug: bool,
        vcfs,
        contig: str,
        prefix: Optional[str] = None,
        min_vaf: float = 0.0,
        output_dir: str = ".",
        keep: bool = False,
        vcfanno_config: Optional[str] = None,
        report_config: Optional[str] = None,
    ) -> None:
        self.debug = debug
        self.vcfs = vcfs[0]
        self.contig = contig
        self.prefix = prefix
        self.min_vaf = min_vaf
        self.output_dir = output_dir
        self.keep = keep
        self.vcfanno_base_path = None
        self.vcfanno_config = vcfanno_config
        self.report_config = report_config

        self.run()

    def run(self) -> None:
        """
        Runs mity report.
        """
        if self.debug:
            logger.setLevel(logging.DEBUG)
            logger.debug("Entered debug mode.")
        else:
            logger.setLevel(logging.INFO)

        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

        # self.vcfs is of type str if there is only one vcf input, but list[str]
        # so we standardise everything to be a list[str]
        if isinstance(self.vcfs, str):
            self.vcfs = [self.vcfs]

        if self.prefix is None:
            self.prefix = MityUtil.make_prefix(self.vcfs[0])

        config_path = os.path.join(MityUtil.get_mity_dir(), "config")
        if self.vcfanno_config is None:
            self.vcfanno_base_path = MityUtil.get_mity_dir()
            match self.contig:
                case "MT":
                    self.vcfanno_config = os.path.join(
                        config_path, "vcfanno-config-mt.toml"
                    )
                case "chrM":
                    self.vcfanno_config = os.path.join(
                        config_path, "vcfanno-config-chrm.toml"
                    )

        if self.report_config is None:
            self.report_config = os.path.join(config_path, "report-config.yaml")

        xlsx_name = os.path.join(
            self.output_dir, self.prefix + ".annotated_variants.xlsx"
        )
        with pandas.ExcelWriter(xlsx_name, engine="xlsxwriter") as writer:
            for vcf in self.vcfs:
                single_report = SingleReport(
                    vcf_path=vcf,
                    min_vaf=self.min_vaf,
                    keep=self.keep,
                    vcfanno_base_path=self.vcfanno_base_path,
                    vcfanno_config=self.vcfanno_config,
                    report_config=self.report_config,
                )
                df = single_report.get_df()

                sheet_name = vcf.replace(".vcf.gz", "").split("/")[-1]
                if len(sheet_name) > 31:
                    logging.info(
                        "sheet_name: %s was too long and was automatically shortened",
                        sheet_name,
                    )
                sheet_name = sheet_name[:31]
                df.to_excel(writer, sheet_name=sheet_name, index=False)
